-- MobArena bot scripts
-- Coded by KirConjurer
-- Import: unitUtil.lua, mobarena.lua

-- These are taken from GAME.EXE (0x59BF18)
local botNames = {
"Dweezle",
"Glork",
"Floogle",
"Goombah",
"Kraun",
"Kloog",
"Zurg",
"Darg",
"Arfingle",
"Buurl",
"Gurgin",
"Grok",
"Hurlong",
"Luric",
"Lupis",
"Mallik",
"Thrall",
"Norwood",
"Nulik",
"Orin",
"Olaf",
"Orguk",
"Pervis",
"Paavik",
"Xevin",
"Xurcon",
"Markoan",
"Yuric",
"Yoovis",
"Yalek",
"Zivik" }

local assignedBotNames = { }

local teamBaseAlertTime = { 0, 0 }

local function getRandomBotName()
    return botNames[math.random(1, #botNames)]
end

local function sendBotChatter(bot, msg)
    sendChat(-1, msg, servNetCode(bot), false)
end

local function mobNoRunaway(unit)
    local uc = getPtrPtr(unit, 0x2EC)
    local fl = getPtrInt(uc, 0x5A0)
    if (bitAnd(fl, 0x20) == 0x20) then
        setPtrInt(uc, 0x5A0, bitXor(fl, 0x20)) -- CAN_CAST_SPELLS
    end
    setPtrFloat(uc, 0x54C, 0) -- flee distance
    setPtrFloat(uc, 0x538, 0) -- health coeff
    return unit
end

local function botDieHandler(bot, killer)
    unitDecay(bot, 20)
    if (killer) then
        if (unitIsPlayer(killer)) then
            local msg = assignedBotNames[bot]
            if (msg) then
                assignedBotNames[bot] = nil
                msg = msg .. " was slain by player " .. playerInfo(killer).name
                sendBotChatter(bot, msg)
            end -- this is possible?
        end
    end
end

local function alertBaseUnderAttack(team)
    teamBaseAlertTime[team] = getFrameCounter()
end

MobArena.BaseDamagedHandler = alertBaseUnderAttack

local function isBaseUnderAttack(team)
    local atime = getFrameCounter() - teamBaseAlertTime[team]
    if (atime < utilSecToFrames(1, 1)) then return true end
    return false
end

local function getNearbyCreatures(bot)
    local result = { }
    local searchRange = 350
    local x, y = unitPos(bot)
    local pick = function(unit) if unitIsMonster(unit) then table.insert(result, unit) end end
    unitGetAround(pick, x - searchRange, y - searchRange, x + searchRange, y + searchRange)
    return result
end

local function getNearestHostileCreature(bot)
    local result = nil
    local x, y = unitPos(bot)
    local visible = getNearbyCreatures(bot)
    if next(visible) == nil then return nil end
    local dist = 40000
    local temp = nil
    local tdist = 0
    for i = 1, #visible do
        temp = visible[i]
        tdist = distance(x, y, unitPos(temp))
        if (unitCanAttack(bot, temp) and unitCanSeeUnit(bot, temp) and tdist < dist) then
            dist = tdist
            result = temp
        end -- sort by distance
    end
    return result
end

local function countNearbyHostiles(bot)
    local result = 0
    local visible = getNearbyCreatures(bot)
    if next(visible) ~= nil then
        local temp = nil
        for i = 1, #visible do
            temp = visible[i]
            if unitCanAttack(bot, temp) then result = result + 1 end
        end
    end
    return result
end

local function monsterGetAttackRange(unit)
    local result = getPtrFloat(unit, 0xB0) + 30
    local defptr = getPtrPtr(getPtrPtr(unit, 0x2EC), 0x1E4)
    result = result + getPtrFloat(defptr, 28)
    return result
end

local function unitGetHPCoeff(unit)
    return unitGetHPSafe(unit) / unitGetMaxHP(unit)
end

local function monsterIsWounded(unit)
    return unitGetHPCoeff(unit) < 0.3
end

local function swapTeams(team)
    if team == 1 then return 2 end
    if team == 2 then return 1 end
    return 0
end

MobArena.BotSkill = 2 -- Normal

local function botThrowAttack()
    local t = 4 - MobArena.BotSkill
    if MobArena.BotSkill > 3 then t = 0 end
    return math.random(0, t) == 0
end

local function botRandomizeAim()
    local e = 0
    if MobArena.BotSkill < 2 then
        e = math.random(0, 4) - 2
    end
    return e
end

local function botGetReactDelay()
    local t = 4 - MobArena.BotSkill
    if MobArena.BotSkill > 3 then t = 0 end
    return t
end

function spawnTrollBot(team)
    local x, y = MobArena.GetSpawnCoords(team)
    local mob = createObject("Troll", x, y)
    -- ||| Setting variables
    local currentGoal = 0 -- Goals: 0 - locate/attack Statue, 1 - fight, 2 - run away, 3 - defend base
    local attackRange = monsterGetAttackRange(mob) -- this is max range in which bot will try to attack enemies
    local minHarpoonRange = 150 -- Harpoon min range
    local maxHarpoonRange = 300 -- Harpoon max range
    local maxStinkRange = 100 -- Stink toggle range
    local stinking = false -- Is Stink active (causes loss of HP but damages enemies nearby)
    local skipCycles = 0 -- 1 cycle is 10 frames, 1 second is 30 frames by default
    local attackReady = true -- Bot is ready to attack
    local attackAnimLength = 60 -- Attack animation length in frames
    local harpoonReady = true -- Harpoon is ready to use
    local locatedStatuePath = false -- Located path to enemy Statue
    local enemyBaseX, enemyBaseY -- Enemy base coords
    enemyBaseX = MobArena.BaseLocations[swapTeams(team)][1]
    enemyBaseY = MobArena.BaseLocations[swapTeams(team)][2]
    local allyBaseX, allyBaseY -- Friendly base coords
    allyBaseX = MobArena.BaseLocations[team][1]
    allyBaseY = MobArena.BaseLocations[team][2]
    -- ||| Spawning alg continues
    unitSetAgression(mob, 0)
    mobNoRunaway(mob)
    unitSetMaxHealth(mob, 80)
    unitSetTeam(mob, team)
    buffApply(mob, 23, utilSecToFrames(4, 1)) -- ENCHANT_INVULNERABLE
    
    local botSwitchGoal = function(n)
        locatedStatuePath = false
        unitActionPop(mob)
        currentGoal = n
        skipCycles = botGetReactDelay()
    end
    
    local updateBotLogic -- why not? every bot will have different logic. and this is much easier than anything else
    updateBotLogic = function()
        x, y = unitPos(mob)
        
        if unitIsDead(mob) then return true
        else setTimeout(updateBotLogic, 10) end
        
        if (not assignedBotNames[mob]) then
            assignedBotNames[mob] = getRandomBotName()
            sendBotChatter(mob, assignedBotNames[mob] .. ", Troll is here!")
        end
        
        if (stinking) then
            if monsterIsWounded(mob) then
                stinking = false
                return
            end
            MobArena.Troll.Stink(mob, mob)
        end
        
        if mobCanMove(mob) and skipCycles == 0 then -- are we able to do anything?
            -- decide what to do depending on current goal
            if currentGoal == 0 then -- Attack Enemy Base
                local enemy = getNearestHostileCreature(mob)
                if (enemy) then -- switch goal to fighting
                    botSwitchGoal(1)
                else -- no enemies?
                    if (distance(x, y, enemyBaseX, enemyBaseY) < attackRange and attackReady) then
                        MobArena.Troll.Attack(mob, enemyBaseX, enemyBaseY)
                        attackReady = false
                        setTimeout(function() attackReady = true end, attackAnimLength)
                    elseif isBaseUnderAttack(team) then
                        botSwitchGoal(3) -- allied base under attack - defend
                    elseif not locatedStatuePath then
                        unitActionPop(mob)
                        setActionTargeted(mob, 7, enemyBaseX, enemyBaseY) -- ACTION_MOVE_TO
                        skipCycles = 1
                        locatedStatuePath = true
                    end
                end
            elseif currentGoal == 1 then -- Fight
                local enemy = getNearestHostileCreature(mob)
                if (enemy) then
                    local ex, ey = unitPos(enemy)
                    local distt = distance(x, y, ex, ey)
                    
                    if (distt < maxStinkRange and not unitIsPoisonImmune(enemy)) then
                        if (not stinking) then soundMake(644, x, y) end
                        stinking = true
                    else
                        stinking = false
                    end -- toggles stinking, aware of enemy range and poison immunity
                    
                    --[[ when bot spots an enemy it will basically do this:
                    1. Enemy is standing close and is not poison immune? = Activate Stink (if not already)
                    2. Enemy is in melee attack range? = Attack, skip this and next cycle
                    3. Enemy is too far to attack but is in harpoon range? = Use Harpoon, skip this cycle
                    4. Else just try to chase enemy, skip next cycle.
                    ]]--
                    if (distt < attackRange and attackReady) then
                        -- enemy in melee attack range? okay
                        MobArena.Troll.Attack(mob, ex, ey)
                        skipCycles = 2 -- wait for attack to finish
                        attackReady = false
                        setTimeout(function() attackReady = true end, attackAnimLength)
                    elseif (distt < maxHarpoonRange and distt > minHarpoonRange and harpoonReady) then
                        -- use harpoon if enemy is going to run
                        ex = ex + botRandomizeAim()
                        ey = ey + botRandomizeAim()
                        MobArena.Troll.Harpoon(mob, ex, ey)
                        skipCycles = 1
                        harpoonReady = false
                        setTimeout(function() harpoonReady = true end, utilSecToFrames(8, 1))
                    elseif distt > attackRange then
                        -- enemy out of range - chase 
                        MobArena_CreatureMoveOld(mob, unitPos(enemy))
                        skipCycles = 1
                    end
                else -- if no enemies left, go for main goal
                    botSwitchGoal(0)
                    stinking = false
                end
            elseif currentGoal == 3 then -- Defend Base
                local enemy = getNearestHostileCreature(mob)
                if (enemy) then
                    botSwitchGoal(1) -- enemy spotted, fight
                elseif not locatedStatuePath then
                    unitActionPop(mob)
                    setActionTargeted(mob, 7, allyBaseX, allyBaseY) -- ACTION_MOVE_TO
                    skipCycles = 1 -- make way to allied base
                    locatedStatuePath = true
                elseif distance(allyBaseX, allyBaseY, unitPos(mob)) < attackRange then
                    botSwitchGoal(0) -- base is safe (no enemies spotted), restore main goal 
                end
            end
        end
        if skipCycles > 0 then skipCycles = skipCycles - 1 end
        return true
    end
    
    setTimeout(updateBotLogic, utilSecToFrames(1, 1)) -- bot will idle for 1 second after spawning
    unitOnDie(mob, botDieHandler)
    return mob
end

function spawnGargoyleBot(team)
    local x, y = MobArena.GetSpawnCoords(team)
    local mob = createObject("EvilCherub", x, y)
    local currentGoal = 0 -- Goals: 0 - locate/attack Statue, 1 - fight, 2 - run away, 3 - defend base
    local attackRange = 200
    local retreatRange = 40
    local deathShotReady = true
    local powerShotReady = true
    local skipCycles = 0
    local attackReady = true
    local attackAnimLength = 25
    local locatedStatuePath = false
    local enemyBaseX, enemyBaseY
    enemyBaseX = MobArena.BaseLocations[swapTeams(team)][1]
    enemyBaseY = MobArena.BaseLocations[swapTeams(team)][2]
    local allyBaseX, allyBaseY
    allyBaseX = MobArena.BaseLocations[team][1]
    allyBaseY = MobArena.BaseLocations[team][2]
    unitSetAgression(mob, 0)
    mobNoRunaway(mob)
    unitSetMaxHealth(mob, 40)
    unitSetTeam(mob, team)
    unitNoPoisonResist(mob)
    buffApply(mob, 23, utilSecToFrames(4, 1)) -- ENCHANT_INVULNERABLE
    
    local botSwitchGoal = function(n)
        locatedStatuePath = false
        unitActionPop(mob)
        currentGoal = n
        skipCycles = botGetReactDelay()
    end
    
    local updateBotLogic
    updateBotLogic = function()
        x, y = unitPos(mob)
        
        if unitIsDead(mob) then return true
        else setTimeout(updateBotLogic, 10) end
        
        if (not assignedBotNames[mob]) then
            assignedBotNames[mob] = getRandomBotName()
            sendBotChatter(mob, assignedBotNames[mob] .. ", Gargoyle is here!")
        end
        
        if mobCanMove(mob) and skipCycles == 0 then
            -- decide what to do depending on current goal
            if currentGoal == 0 then
                local enemy = getNearestHostileCreature(mob)
                if (enemy) then -- switch goal to fighting
                    botSwitchGoal(1)
                else -- go to base
                    if (distance(x, y, enemyBaseX, enemyBaseY) < attackRange and attackReady) then
                        MobArena.Gargoyle.Attack(mob, enemyBaseX, enemyBaseY)
                        attackReady = false
                        setTimeout(function() attackReady = true end, attackAnimLength)
                    elseif isBaseUnderAttack(team) then
                        botSwitchGoal(3) -- allied base under attack - defend
                    elseif not locatedStatuePath then
                        unitActionPop(mob)
                        setActionTargeted(mob, 7, enemyBaseX, enemyBaseY) -- ACTION_MOVE_TO
                        skipCycles = 1
                        locatedStatuePath = true
                    end
                end
            elseif currentGoal == 1 then
                local enemy = getNearestHostileCreature(mob)
                if (enemy) then
                    local ex, ey = unitPos(enemy)
                    local distt = distance(x, y, ex, ey)
                    
                    if distt < retreatRange then
                        -- keep distance
                        if (powerShotReady) then
                            ex = ex + botRandomizeAim()
                            ey = ey + botRandomizeAim()
                            MobArena.Gargoyle.PowerShot(mob, ex, ey)
                            powerShotReady = false
                            setTimeout(function() powerShotReady = true end, utilSecToFrames(6, 1))
                        end -- first try to stun enemy
                        unitActionPop(mob)
                        unitSetAction(mob, 29) -- ACTION_RANDOM_WALK
                        skipCycles = 3
                    elseif distt < attackRange then
                        -- Prioritize abilities over normal attacks
                        if (deathShotReady) then
                            ex = ex + botRandomizeAim()
                            ey = ey + botRandomizeAim()
                            MobArena.Gargoyle.DeathShot(mob, ex, ey)
                            deathShotReady = false
                            setTimeout(function() deathShotReady = true end, utilSecToFrames(4, 1))
                        elseif (attackReady) then
                            ex = ex + botRandomizeAim()
                            ey = ey + botRandomizeAim()
                            MobArena.Gargoyle.Attack(mob, ex, ey)
                            skipCycles = 1 -- wait for attack to finish
                            attackReady = false
                            setTimeout(function() attackReady = true end, attackAnimLength)
                        end
                    elseif distt > attackRange then
                        -- try to chase enemy
                        MobArena_CreatureMoveOld(mob, unitPos(enemy))
                        skipCycles = 1
                    end
                else botSwitchGoal(0) end
            elseif currentGoal == 3 then -- Defend Base
                local enemy = getNearestHostileCreature(mob)
                if (enemy) then
                    botSwitchGoal(1) -- enemy spotted, fight
                elseif not locatedStatuePath then
                    unitActionPop(mob)
                    setActionTargeted(mob, 7, allyBaseX, allyBaseY) -- ACTION_MOVE_TO
                    skipCycles = 1 -- make way to allied base
                    locatedStatuePath = true
                elseif distance(allyBaseX, allyBaseY, unitPos(mob)) < attackRange then
                    botSwitchGoal(0) -- base is safe (no enemies spotted), restore main goal 
                end
            end
        end
        if skipCycles > 0 then skipCycles = skipCycles - 1 end
        return true
    end
    
    setTimeout(updateBotLogic, utilSecToFrames(1, 1)) -- bot will idle for 1 second after spawning
    unitOnDie(mob, botDieHandler)
    return mob
end

function spawnNecromancerBot(team)
    local x, y = MobArena.GetSpawnCoords(team)
    local mob = createObject("Necromancer", x, y)
    local currentGoal = 0 -- Goals: 0 - locate/attack Statue, 1 - fight, 2 - run away, 3 - defend base
    local attackRange = 200
    local retreatRange = 50
    local chillTouchReady = true
    local deepFreezeReady = true
    local skipCycles = 0
    local attackReady = true
    local attackAnimLength = 24
    local locatedStatuePath = false
    local enemyBaseX, enemyBaseY
    enemyBaseX = MobArena.BaseLocations[swapTeams(team)][1]
    enemyBaseY = MobArena.BaseLocations[swapTeams(team)][2]
    local allyBaseX, allyBaseY
    allyBaseX = MobArena.BaseLocations[team][1]
    allyBaseY = MobArena.BaseLocations[team][2]
    unitSetAgression(mob, 0)
    mobNoRunaway(mob)
    unitSetMaxHealth(mob, 50)
    unitSetTeam(mob, team)
    unitNoPoisonResist(mob)
    buffApply(mob, 23, utilSecToFrames(4, 1)) -- ENCHANT_INVULNERABLE
    
    local botSwitchGoal = function(n)
        locatedStatuePath = false
        unitActionPop(mob)
        currentGoal = n
        skipCycles = botGetReactDelay()
    end
    
    local updateBotLogic
    updateBotLogic = function()
        x, y = unitPos(mob)
        
        if unitIsDead(mob) then return true 
        else setTimeout(updateBotLogic, 10) end
        
        if (not assignedBotNames[mob]) then
            assignedBotNames[mob] = getRandomBotName()
            sendBotChatter(mob, assignedBotNames[mob] .. ", Necromancer is here!")
        end
        
        if mobCanMove(mob) and skipCycles == 0 then 
            -- decide what to do depending on current goal
            if currentGoal == 0 then
                local enemy = getNearestHostileCreature(mob)
                if (enemy) then -- switch goal to fighting
                    botSwitchGoal(1)
                else -- go to base
                    if (distance(x, y, enemyBaseX, enemyBaseY) < attackRange and attackReady) then
                        MobArena.Necromancer.Attack(mob, enemyBaseX, enemyBaseY)
                        attackReady = false
                        setTimeout(function() attackReady = true end, attackAnimLength)
                    elseif isBaseUnderAttack(team) then
                        botSwitchGoal(3) -- allied base under attack - defend
                    elseif not locatedStatuePath then
                        unitActionPop(mob)
                        setActionTargeted(mob, 7, enemyBaseX, enemyBaseY) -- ACTION_MOVE_TO
                        skipCycles = 1
                        locatedStatuePath = true
                    end
                end
            elseif currentGoal == 1 then
                local enemy = getNearestHostileCreature(mob)
                if (enemy) then
                    local ex, ey = unitPos(enemy)
                    local distt = distance(x, y, ex, ey)
                    
                    --if monsterIsWounded(mob) then botSwitchGoal(2) end -- retreat, regen health
                    
                    if distt < retreatRange then
                        -- keep distance
                        
                        unitActionPop(mob)
                        unitSetAction(mob, 29) -- ACTION_RANDOM_WALK
                        skipCycles = 3
                    elseif (chillTouchReady and distt < 375) then
                        if botThrowAttack() then
                            MobArena.Necromancer.ChillTouch(mob, enemy)
                            chillTouchReady = false
                            setTimeout(function() chillTouchReady = true end, utilSecToFrames(8, 1))
                        end
                        skipCycles = 1
                    elseif (deepFreezeReady and distt < 200 and countNearbyHostiles(mob) > 2) then
                        if botThrowAttack() then
                            MobArena.Necromancer.DeepFreeze(mob)
                            deepFreezeReady = false
                            setTimeout(function() deepFreezeReady = true end, utilSecToFrames(14, 1))
                        end
                        skipCycles = 1
                    elseif distt < attackRange then
                        -- Prioritize abilities over normal attacks
                        if (attackReady) then
                            ex = ex + botRandomizeAim()
                            ey = ey + botRandomizeAim()
                            MobArena.Necromancer.Attack(mob, ex, ey)
                            skipCycles = 1 -- wait for attack to finish
                            attackReady = false
                            setTimeout(function() attackReady = true end, attackAnimLength)
                        end
                    elseif distt > attackRange then
                        -- try to chase enemy
                        MobArena_CreatureMoveOld(mob, unitPos(enemy))
                        skipCycles = 1
                    end
                else botSwitchGoal(0) end
            elseif currentGoal == 2 then
                if not monsterIsWounded(mob) then botSwitchGoal(0) end -- go for main goal
                
                local enemy = getNearestHostileCreature(mob)
                if (enemy) then
                    unitActionPop(mob)
                    creatureActionRoam(mob)
                    skipCycles = 3
                end -- mob will simply try to regen health ignoring enemies
            elseif currentGoal == 3 then -- Defend Base
                local enemy = getNearestHostileCreature(mob)
                if (enemy) then
                    botSwitchGoal(1) -- enemy spotted, fight
                elseif not locatedStatuePath then
                    unitActionPop(mob)
                    setActionTargeted(mob, 7, allyBaseX, allyBaseY) -- ACTION_MOVE_TO
                    skipCycles = 1 -- make way to allied base
                    locatedStatuePath = true
                elseif distance(allyBaseX, allyBaseY, unitPos(mob)) < attackRange then
                    botSwitchGoal(0) -- base is safe (no enemies spotted), restore main goal 
                end
            end
        end
        if skipCycles > 0 then skipCycles = skipCycles - 1 end
        return true
    end
    
    setTimeout(updateBotLogic, utilSecToFrames(1, 1)) -- bot will idle for 1 second after spawning
    unitOnDie(mob, botDieHandler)
    return mob
end

local allBotTypes = { spawnTrollBot, spawnGargoyleBot, spawnNecromancerBot }

MobArena.SpawnRandomBots = function()
    local botSpawn = nil
    for i = 1, 3 do
        botSpawn = allBotTypes[math.random(1, #allBotTypes)]
        botSpawn(1)
    end
    for i = 1, 3 do
        botSpawn = allBotTypes[math.random(1, #allBotTypes)]
        botSpawn(2)
    end
end
        