-- Monster-related scripts (part of MobArena)
-- Coded by KirConjurer

function setActionTargeted(mob, act, x, y)
    local t = unitSetAction(mob, act)
    setPtrFloat(t, 4, x)
    setPtrFloat(t, 8, y)
    setPtrInt(t, 12, 0)
    return t
end

--[[function setActionFight(mob, targ)
 local t = unitSetAction(mob, 15) -- ACTION_FIGHT
 local x,y = unitPos(targ)
 setPtrFloat(t, 4, x)
 setPtrFloat(t, 8, y)
 setPtrInt(t, 12, getFrameCounter() + 10)
 return t
end]]--

local function setActionCastAt(mob, targ, id)
    local t = unitSetAction(mob, 18) -- ACTION_CAST_SPELL_ON_OBJECT
    setPtrInt(t, 4, id)
    setPtrInt(t, 8, 0)
    setPtrPtr(t, 12, targ)
    return t
end

local function setActionCastTo(mob, x, y, id)
    local t = unitSetAction(mob, 19) -- ACTION_CAST_SPELL_ON_LOCATION
    setPtrInt(t, 4, id)
    setPtrFloat(t, 12, x)
    setPtrFloat(t, 16, y)
    return t
end

function projectileSetup(proj, x, y, fx, fy)
    local bx, by = unitPos(proj)
    local vx = x - bx
    local vy = y - by
    local vlen = math.sqrt(vx ^ 2 + vy ^ 2)
    vx = vx / vlen * 10
    vy = vy / vlen * 10
    if vx == vx and vy == vy then
        vx = vx + bx
        vy = vy + by
        if mapTraceRay(bx, by, vx, vy, 5) then
            unitMove(proj, vx, vy)
            --setPtrInt(proj, 0x10, bitOr(unitFlags(proj), 0x21000400))
            --ph(unitFlags(proj))
            setPtrFloat(proj, 0x50, fx)
            setPtrFloat(proj, 0x54, fy)
            unitSetLookAngle(proj, directGet(x, y, vx, vy))
            return true
        end
    end
    return false
end

local function trollStinkEffect(mob, parent)
    unitDamage(mob, math.random(1, 3)) -- pure damage
    local px, py = unitPos(mob)
    local cloud = createToxicCloud(px, py, 10)
    if not parent then parent = mob end
    unitSetParent(cloud, parent) -- set parent
    return cloud
end

local function archmageFlame(mob, x, y)
    buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
    soundMake(9, unitPos(mob)) -- CleansingFlameCast
    local flame = createObject("Flame", x, y)
    unitSetParent(flame, mob)
    unitDecay(flame, 60) -- 60 damage if aimed correctly
    unitActionPop(mob) setActionCastTo(mob, x, y, 0) -- animation
    return flame
end

function MobArena_CreatureMoveDefault(mob, x, y)
    local mx, my = unitPos(mob)
    if (MobArena.UseOldMS) then
        MobArena_CreatureMoveOld(mob, x, y)
        return
    end
    if (x > 0 and y > 0 and x < 5887 and y < 5887 and distance(x, y, mx, my) > 10) then
        unitSetLookAngle(mob, directGet(x, y, mx, my))
        unitActionPop(mob)
        setActionTargeted(mob, 9, x, y) -- ACTION_DODGE
        
        local ss = monsterGetSoundset(mob)
        local move = getPtrInt(ss, 72)
        if (move) then
            soundMake(move, mx, my)
        end -- look 0x534030
    end
end

function MobArena_CreatureMoveOld(mob, x, y) -- pathfinding
    local mx, my = unitPos(mob)
    if (mapTraceRay(mx, my, x, y)) then
        unitActionPop(mob)
        setActionTargeted(mob, 7, x, y) -- ACTION_MOVE_TO
    end
end

local trollLines = {
    "Nice catch!",
    "Welcome!",
    "Get over here!",
    "What a surprise!"
}

local function fakeHarpoonCollide(harp, targ)
    unitOnCollide(harp, nil) -- alternative fix
    local owner = unitGetParent(harp) 
    -- validation block
    if (not objectExists(owner)) then return true end
    if (not objectExists(targ)) then return true end
    -- end validation
    local mx, my = unitPos(owner)
    local tx, ty, vx, vy, line
    if (unitIsMonster(targ)) then
        tx, ty = unitPos(targ)
        if isMagicVulnerable(targ) then unitDamage(targ, 10) end -- don't damage invulnerable units
        -- calculate vector
        vx = mx - tx
        vy = my - ty
        vx = vx / 1.92
        vy = vy / 1.92
        buffApply(targ, 5, 6) -- stun for 6 frames
        unitNoCollide(targ, false) -- makes trees/windows/etc passable
        unitFlAirborne(targ, false)
        setTimeout(function() if objectExists(targ) then
            unitNoCollide(targ, true)
            unitFlAirborne(targ, true)
            soundMake(998, unitPos(targ))
        end end, 6) -- remove flag after 6 frames
        if vx == vx and vy == vy then
            unitSpeed(targ, vx, vy) -- push target
        end -- test for -1.#IND
        soundMake(999, unitPos(targ)) -- reel sound
        unitDelete(harp) -- delete harpoon
        line = math.random(1, #trollLines) -- random chatter
        sendChat(-1, trollLines[line], servNetCode(owner), false)
    end
    return true
end

local function trollLaunchHarpoon(mob, x, y)
    buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
    soundMake(177, unitPos(mob)) -- harpoon sound
    local harp = createObject('MercArcherArrow', unitPos(mob)) -- TODO: fix WeakArcherArrow (0x4EB490)
    unitSetParent(harp, mob)
    local angle = directGet(x, y, unitPos(mob))
    unitSetLookAngle(mob, angle)
    unitSetLookAngle(harp, angle)
    unitOnCollide(harp, fakeHarpoonCollide)
    local mx, my = unitPos(mob)
    local vx, vy
    vx = x - mx
    vy = y - my
    local vlen = math.sqrt(vx ^ 2 + vy ^ 2)
    vx = vx / vlen * 86
    vy = vy / vlen * 86
    if vx == vx and vy == vy then
        unitSpeed(harp, vx, vy)
        setTimeout(function() if objectExists(harp) then unitDelete(harp) end end, 30)--unitOnCollide(harp, nil) end, 39) fix 0x53AC10
    else
        unitDelete(harp)
    end -- test for -1.#IND
    return harp
end

local function gargPowerShotCollide(shot, targ)
    unitOnCollide(shot, nil)
    local owner = unitGetParent(shot)
    -- validation block
    if (not objectExists(owner)) then return true end
    if (not objectExists(targ)) then return true end
    -- end validation
    if (unitIsMonster(targ) and not unitTestBuff(targ, 23)) then
        if (not unitTestBuff(targ, 19)) then
            buffApply(targ, 5, math.random(utilSecToFrames(2, 1), utilSecToFrames(3, 1))) -- stun 2-3 sec
        end
        unitDamage(targ, math.random(10, 20))
        sendChat(-1, 'Hit!', servNetCode(owner), false)
        unitDelete(shot)
    end
    return true
end

local function gargCreatePowerShot(mob, x, y)
    buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
    local shot = createObject('CherubArrow', unitPos(mob))
    unitSetParent(shot, mob)
    unitSetLookAngle(shot, unitGetLookAngle(mob))
    unitOnCollide(shot, gargPowerShotCollide)
    local mx, my = unitPos(mob)
    local vx, vy
    vx = x - mx
    vy = y - my
    local vlen = math.sqrt(vx ^ 2 + vy ^ 2)
    vx = vx / vlen * 160
    vy = vy / vlen * 160
    if vx == vx and vy == vy then
        unitSpeed(shot, vx, vy)
        setTimeout(function() unitOnCollide(shot, nil) end, 30)
    else
        unitDelete(shot)
    end -- test for -1.#IND
    soundMake(886, unitPos(mob))
    return shot
end

local function gargDeathShotCollide(shot, targ)
    unitOnCollide(shot, nil)
    local owner = unitGetParent(shot)
    -- validation block
    if (not objectExists(owner)) then return true end
    if (not objectExists(targ)) then return true end
    -- end validation
    if (unitIsMonster(targ) and not unitTestBuff(targ, 23)) then
        local collData = getPtrPtr(shot, 0x2BC)
        local dmg = getPtrInt(collData, 0)
        unitDamage(targ, dmg)
        unitAddHP(owner, dmg) -- 100% vampirism
        local ox, oy = unitPos(owner)
        for i = 1, 3 do netRayFx(0x91, ox, oy, unitPos(targ)) end
        soundMake(5, unitPos(targ))
        unitDelete(shot)
    end
    return true
end

local function gargCreateDeathShot(mob, x, y)
    buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
    local shot = createObject('CherubArrow', unitPos(mob))
    unitSetParent(shot, mob)
    unitSetLookAngle(shot, unitGetLookAngle(mob))
    unitOnCollide(shot, gargDeathShotCollide)
    local mx, my = unitPos(mob)
    local vx, vy
    vx = x - mx
    vy = y - my
    local vlen = math.sqrt(vx ^ 2 + vy ^ 2)
    vx = vx / vlen * 150
    vy = vy / vlen * 150
    if vx == vx and vy == vy then
        unitSpeed(shot, vx, vy)
        setTimeout(function() unitOnCollide(shot, nil) end, 30)
    else
        unitDelete(shot)
    end -- test for -1.#IND
    local dmg = math.ceil(unitGetHPSafe(mob) * 0.15)
    if dmg < 1 then dmg = 1 end
    unitDamage(mob, dmg)
    dmg = dmg * 2 -- (15% current hp) x2
    arrowSetDamage(shot, dmg)
    soundMake(575, unitPos(mob))
    soundMake(896, unitPos(mob))
    return shot
end

local function archmageMassHeal(mob)
    local mx, my = unitPos(mob)
    local d = 160
    netPointFx(0xA0, mx, my)
    soundMake(773, mx, my)
    local fn = function(p) if unitIsMonster(p) and (not unitCanAttack(mob, p)) then
        unitAddHP(p, 20)
        soundMake(754, unitPos(p))
        if (not unitTestBuff(p, 23)) then
            buffApply(p, 23, utilSecToFrames(1, 1)) -- ENCHANT_INVULNERABLE
        end
    end end
    setTimeout(function()
        unitGetAround(fn, mx - d, my - d, mx + d, my + d)
    end, utilSecToFrames(1, 0.1))
end

function distance(x1, y1, x2, y2)
    local dx = x1 - x2
    local dy = y1 - y2
    return math.sqrt(dx * dx + dy * dy)
end

local function necroApplyDF(targ)
    if isMagicVulnerable(targ) then
        unitDamage(targ, math.random(10, 20)) -- pure damage
        buffApply(targ, 25, math.random(utilSecToFrames(2, 1), utilSecToFrames(3, 1))) -- 2-3 sec ENCHANT_FREEZE
    end -- don't apply to invulnerables
    netPointFx(0x81, unitPos(targ))
end

local function necroDeepFreeze(mob)
    buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
    local x, y = unitPos(mob)
    unitActionPop(mob) setActionCastAt(mob, mob, 0) -- animation
    local fn = function(p) if (unitIsMonster(p) and unitCanAttack(mob, p)) then necroApplyDF(p) end end
    local d = 200 -- active radius
    setTimeout(function()
        if not unitIsDead(mob) then
            if mobCanMove(mob) then
                unitGetAround(fn, x - d, y - d, x + d, y + d)
                netPointFx(0xA3, x, y)
                soundMake(913, x, y)
                sendChat(-1, "Frozen forever!", servNetCode(mob), false)
            else
                sendChat(-1, "Arrgh!", servNetCode(mob), false)
            end
        end
    end, utilSecToFrames(1, 0.5)) -- cast can be interrupted
    return true
end

local function dryadNatureWrath(mob, targ)
    buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
    local mx, my = unitPos(mob)
    local tx, ty = unitPos(targ)
    local damage = 20
    local secondary = {} -- secondary targets
    local secTargs = 0
    local d = 160 -- half of 320
    local extraDmg = (unitGetMaxHP(mob) - unitGetHPSafe(mob)) * 1.25
    local maxBounce = math.ceil(extraDmg / 10) + 1 -- max 6 bounces when 1 hp left
    local sec
    damage = damage + math.ceil(extraDmg) -- -1 hp|+1.25 dmg
    netGreenLightning(mx, my, tx, ty, 8)
    unitDamage(targ, damage)
    local fn = function(p)
        if (unitIsMonster(p) and unitCanAttack(mob, p)) then
            secTargs = secTargs + 1
            if (secTargs <= maxBounce) then secondary[secTargs] = p end
        end
    end -- get nearby creatures, up to limit
    unitGetAround(fn, tx - d, ty - d, tx + d, ty + d)
    for i = 1, secTargs do
        sec = secondary[i]
        if (sec and sec ~= targ) then
            mx = tx
            my = ty
            tx, ty = unitPos(sec)
            netGreenLightning(mx, my, tx, ty, 8)
            if isMagicVulnerable(sec) then unitDamage(sec, damage) end
        end -- don't hit main target twice
    end
    soundMake(221, mx, my)
end

local function createWaspNest(x, y, hp)
    local nest = createObject("WaspNest", x, y)
    local flags = unitFlags(nest)
    flags = bitXor(flags, 0x10) -- this allows melee attacking
    --flags = bitOr(flags, 0x800) -- allow allies to pass through
    setPtrInt(nest, 0x10, flags)
    unitHP(nest, hp)
    return nest
end

local function lichColdFlameCollide(obj, coll)
    local owner = unitGetParent(obj)
    -- validation block
    if (not objectExists(owner)) then return true end
    if (not objectExists(coll)) then return true end
    -- end validation
    if (unitIsMonster(coll) and isMagicVulnerable(coll)) then
        if (not unitTestBuff(coll, 25) and unitCanAttack(owner, coll)) then
            buffApply(coll, 25, utilSecToFrames(1, 0.4)) -- ENCHANT_FREEZE
        end
    end
    return true
end

local function beholderCastFlash(mob)
    buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
    local mx, my = unitPos(mob)
    netPointFx(0x9A, mx, my)
    soundMake(157, mx, my)
end

local function utilCheckNaN(a, b)
    if a ~= a or b ~= b then -- test for -1.#IND
        a = math.random(0, 20) - 10
        b = math.random(0, 20) - 10
    end
    return a, b
end

local function wizardWhirlwindEffect(targ, mob, wx, wy)
    if (isMagicVulnerable(targ) and unitCanAttack(mob, targ)) then
        local tx, ty = unitPos(targ)
        local power = (120 - distance(wx, wy, tx, ty)) / 20
        power = math.abs(power)
        local vx, vy
        vx = wx - tx
        vy = wy - ty
        local vlen = math.sqrt(vx ^ 2 + vy ^ 2)
        vx = vx / vlen * 3
        vy = vy / vlen * 3
        vx, vy = utilCheckNaN(vx, vy)
        unitSpeed(targ, vx, vy)
        unitDamage(targ, math.ceil(power)) -- pure damage
        if not unitTestBuff(targ, 4) then
            buffApply(targ, 4, utilSecToFrames(1, 1)) -- ENCHANT_SLOWED
        end
    end -- don't apply to invulnerables
end

local function wizardApplyExplosion(targ, x, y)
    local tx, ty = unitPos(targ)
    local power = (150 - distance(x, y, tx, ty)) / 40
    power = math.abs(power)
    local vx, vy
    vx = tx - x
    vy = ty - y
    local vlen = math.sqrt(vx ^ 2 + vy ^ 2)
    vx = vx / vlen * 10
    vy = vy / vlen * 10
    vx, vy = utilCheckNaN(vx, vy)
    unitSpeed(targ, vx, vy)
    if isMagicVulnerable(targ) then
        unitDamage(targ, math.ceil(power * 8))
        local stunDur = math.floor(power * 6) + math.random(1, 5)
        buffApply(targ, 5, stunDur) -- stun up to 1 sec
    end
end

local function wizardExplosionEffect(mob, x, y)
    local fn = function(p) if (unitIsMonster(p) and unitCanAttack(mob, p)) then wizardApplyExplosion(p, x, y) end end
    local d = 100 -- active radius
    unitGetAround(fn, x - d, y - d, x + d, y + d)
    netPointFx(0x85, x, y)
    soundMake(285, x, y)
    return true
end

local function mechFlyerTripleShot(mob, x, y, parent)
    buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
    unitActionPop(mob)
    unitSetLookAngle(mob, directGet(x, y, unitPos(mob)))
    local look = unitGetLookAngle(mob)
    local shot, vx, vy
    for ia = -7, 7, 7 do
        shot = createObject("GolemArrow", unitPos(mob))
        if not parent then parent = mob end
        unitSetParent(shot, parent) -- set parent
        unitSetLookAngle(shot, look)
        vx = utilCosFromDirect(look + ia) * 90
        vy = utilSinFromDirect(look + ia) * 90
        unitSpeed(shot, vx, vy)
        arrowSetDamage(shot, 10) -- increase damage
    end
    soundMake(889, unitPos(mob))
end

local function mechFlyerSelfRepair(mob)
    local mx, my = unitPos(mob)
    soundMake(811, mx, my)
    soundMake(803, mx, my)
    unitAddHP(mob, 50)
end

local function defaultMissileAttack(mob, x, y)
    unitActionPop(mob)
    unitSetLookAngle(mob, directGet(x, y, unitPos(mob)))
    setActionTargeted(mob, 17, x, y) -- ACTION_MISSILE_ATTACK
end

local function defaultMeleeAttack(mob, x, y)
    unitActionPop(mob)
    unitSetLookAngle(mob, directGet(x, y, unitPos(mob)))
    unitSetAction(mob, 16) -- ACTION_MELEE_ATTACK
end

local function ogreMagicResistance(mob)
    local mx, my = unitPos(mob)
    buffApply(mob, 19, utilSecToFrames(5, 1)) -- ENCHANT_PROTECT_FROM_MAGIC, 5 sec
    soundMake(6, mx, my)
    --sendChat(-1, "Wizardry means nothing to me!", servNetCode(mob), false)
end

local function zombiePlaceTrap(mob, parent)
    local trap = createObject("PoisonGasTrap", unitPos(mob))
    if not parent then parent = mob end
    unitSetParent(trap, parent)
    soundMake(837, unitPos(mob))
    return trap
end

local function dryadAttack(mob, x, y, parent)
    buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
    unitActionPop(mob) setActionCastTo(mob, x, y, 0) -- animation
    local mx, my = unitPos(mob)
    local vx, vy
    local shot = createObject("YellowStarShot", mx, my)
    local cdat = getPtrPtr(shot, 0x2BC)
    setPtrInt(cdat, 0, math.random(9, 11)) -- increase damage
    if not parent then parent = mob end
    unitSetParent(shot, parent) -- set parent
    unitSetLookAngle(mob, directGet(x, y, mx, my))
    vx = x - mx
    vy = y - my
    local vlen = math.sqrt(vx ^ 2 + vy ^ 2)
    vx = vx / vlen * 40
    vy = vy / vlen * 40
    if vx == vx and vy == vy then unitSpeed(shot, vx, vy)
    else unitDelete(shot) end
    soundMake(220, unitPos(mob))
end

local function dryadApplyFear(mob, targ)
    buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
    if isMagicVulnerable(targ) then
        buffApply(targ, 11, utilSecToFrames(2, 1)) -- ENCHANT_AFRAID, 2 secs
    end
end

local function archmageAttack(mob, x, y, parent)
    buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
    unitActionPop(mob) setActionCastTo(mob, x, y, 0) -- animation
    local mx, my = unitPos(mob)
    local vx, vy
    local shot = createObject("PitifulFireball", mx, my)
    if not parent then parent = mob end
    unitSetParent(shot, parent) -- set parent
    local look = directGet(x, y, mx, my)
    unitSetLookAngle(mob, look) unitSetLookAngle(shot, look)
    vx = x - mx
    vy = y - my
    local vlen = math.sqrt(vx ^ 2 + vy ^ 2)
    vx = vx / vlen * 40
    vy = vy / vlen * 40
    if vx == vx and vy == vy then unitSpeed(shot, vx, vy)
    else unitDelete(shot) end
    soundMake(218, mx, my)
end

local function archmageCastFF(mob, targ)
    setActionCastAt(mob, targ, 51) -- SPELL_SHIELD
end

local function necromancerAttack(mob, x, y, parent)
    buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
    unitSetLookAngle(mob, directGet(x, y, unitPos(mob)))
    unitActionPop(mob) --setActionCastTo(mob, x, y, 0) -- animation
    local mx, my = unitPos(mob)
    local vx = x - mx
    local vy = y - my
    local vlen = math.sqrt(vx ^ 2 + vy ^ 2)
    vx = vx / vlen * 45
    vy = vy / vlen * 45
    soundMake(907, mx, my)
    local shot = createObject("LightningBolt", mx, my)
    if not parent then parent = mob end
    unitSetParent(shot, parent) -- set parent
    unitSetLookAngle(shot, unitGetLookAngle(mob))
    if vx == vx and vy == vy then unitSpeed(shot, vx, vy)
    else unitDelete(shot) end
end

local function necromancerColdTouch(mob, targ)
    buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
    local x, y = unitPos(targ)
    if isMagicVulnerable(targ) then
        unitDamage(targ, 10) -- ^ check for ENCHANT_INVULNERABLE
        buffApply(targ, 25, utilSecToFrames(3, 1)) -- 3 sec ENCHANT_FREEZE
    end
    unitActionPop(mob) setActionCastTo(mob, x, y, 0) -- animation
    for i = 1, 5 do netRayFx(0x8F, x, y, unitPos(mob)) end
end

local function necromancerRaiseZombie(mob, x, y, parent)
    buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
    --local mx, my = unitPos(mob)
    unitActionPop(mob) setActionCastTo(mob, x, y, 0) -- animation
    --if mapTraceRay(mx, my, x, y, 1) then
    netPointFx(0x8A, x, y)
    local zomb = createObject("Zombie", x, y)
    unitSetMaxHealth(zomb, 50)
    if (parent) then
        unitBecomePet(zomb, parent)
    else
        unitSetParent(zomb, mob)
    end
    setTimeout(function() unitSetAgression(zomb, 0.83) unitSetFollow(zomb, mob) end, 1)
    setTimeout(function() unitDecay(zomb, 20) end, utilSecToFrames(6, 1)) -- lifetime: 6 sec
    return zomb
end

local function shamanAttack(mob, x, y, parent)
    buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
    unitActionPop(mob) setActionCastTo(mob, x, y, 0) -- animation
    local cx, cy = unitPos(mob)
    local missile = createMagicMissile(cx, cy, parent)
    if not projectileSetup(missile, x, y, 0, 0) then unitDelete(missile) end

    soundMake(218, cx, cy)
end

local function shamanSummonMeteor(mob, x, y, parent)
    buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
    unitActionPop(mob) setActionCastTo(mob, x, y, 0) -- animation
    local m = createFallingMeteor(x, y, -10, 32)
    if not parent then parent = mob end
    unitSetParent(m, parent)
    soundMake(85, x, y)
end

local function shamanHealingRitual(mob, targ)
    local hp = unitGetHPSafe(targ)
    local coeff = hp / unitGetMaxHP(targ)
    local heal = unitGetMaxHP(targ) / 4
    if coeff < 0.5 then
        heal = unitGetMaxHP(targ) / 2
    end
    local tx, ty = unitPos(targ)
    unitActionPop(mob) setActionCastTo(mob, tx, ty, 0) -- animation
    heal = math.floor(heal)
    setTimeout(function()
        if (not unitIsDead(mob) and not unitIsDead(targ)) then
            unitAddHP(targ, heal)
            tx, ty = unitPos(targ)
            netGeneratorBreakFX(tx, ty)
            soundMake(754, tx, ty)
        end
    end, utilSecToFrames(1, 0.5))
end

local function shamanApplyConfuse(mob, targ)
    buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
    if isMagicVulnerable(targ) then
        buffApply(targ, 3, utilSecToFrames(3, 1)) -- ENCHANT_CONFUSED, 3 secs
    end
end

local function spiderRemotePoison(mob, targ)
    buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
    if isMagicVulnerable(targ) then
        unitDamage(targ, 10) -- pure damage
    end
    unitPoison(targ, 3) -- poison lvl 3
end

local function spiderCallSpiderlings(mob, parent)
    for n = 1, 2 do
        local spi = createObject("SmallSpider", unitPos(mob))
        unitSetMaxHealth(spi, 10)
        if (parent) then
            unitBecomePet(spi, parent)
        else
            unitSetParent(spi, mob)
        end -- BecomePet works only with players
        setTimeout(function() unitSetAgression(spi, 0.83) unitSetFollow(spi, mob) end, 1)
        setTimeout(function() unitDecay(spi, 20) end, utilSecToFrames(10, 1)) -- lifetime: 10 sec
    end
end

local function spiderShootWeb(mob, parent)
    buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
    local look = unitGetLookAngle(mob)
    local spit, vx, vy
    for ia = -10, 10, 5 do
        spit = createObject("SpiderSpit", unitPos(mob))
        if not parent then parent = mob end
        unitSetParent(spit, parent)
        unitSetLookAngle(spit, look)
        vx = utilCosFromDirect(look + ia) * 45
        vy = utilSinFromDirect(look + ia) * 45
        unitSpeed(spit, vx, vy)
    end
    soundMake(348, unitPos(mob))
end

local function impCloneSelf(mob, parent)
    buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
    local clones = { }
    netPointFx(0x83, unitPos(mob))
    local cHP = unitGetHPSafe(mob) / 2
    if (cHP < 1) then cHP = 1 end
    for n = 1, 5 do
        local imp = createObject("Imp", unitPos(mob))
        table.insert(clones, imp)
        unitSetMaxHealth(imp, cHP) -- half HP of original
        if (parent) then
            unitBecomePet(imp, parent)
        else
            unitSetParent(imp, mob)
        end -- BecomePet works only with players
        setTimeout(function() unitSetAgression(imp, 0) unitSetFollow(imp, mob) end, 1)
        setTimeout(function() unitDecay(imp, 20) end, 300) -- lifetime: 10 sec
    end
    return clones
end

local function impClonesAttack(x, y, clones)
    local imp
    for i = 1, #clones do
        imp = clones[i]
        if not unitIsDead(imp) then
            defaultMissileAttack(imp, x, y)
        end
    end
end

local function impApplyHaste(mob)
    buffApply(mob, 9, utilSecToFrames(3, 1)) -- ENCHANT_HASTED
end

local function bearDoBash(mob)
    buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
    local bashRange = 80
    local mx, my = unitPos(mob)
    soundMake(485, mx, my)
    sendEarthquake(mx, my, 25)
    netPointFx(0x8A, mx, my)
     
    local bearBashSlow = function(targ)
        if (unitIsMonster(targ) and isMagicVulnerable(targ) and unitCanAttack(mob, targ)) then
            buffApply(targ, 4, math.random(utilSecToFrames(1, 1), utilSecToFrames(2, 1))) -- 1-2 sec ENCHANT_SLOWED
            unitDamage(targ, 8) -- pure damage
        end -- don't apply to invulnerables and user
    end
    unitGetAround(bearBashSlow, mx-bashRange, my-bashRange, mx+bashRange, my+bashRange)
end

local function shadeJump(mob, targ)
    buffCancel(mob, 23) -- ENCHANT_INVULNERABLE
    local mx, my = unitPos(mob)
    local x, y = unitPos(targ)
    local vx, vy
    vx = x - mx
    vy = y - my
    vx = vx / 1.92
    vy = vy / 1.92
    unitNoCollide(mob, false)
    unitFlAirborne(mob, false)
    setTimeout(function() if objectExists(mob) then
        unitNoCollide(mob, true)
        unitFlAirborne(mob, true)
    end end, 6) -- remove flag after 6 frames
    if (unitCanAttack(mob, targ)) then
        buffApply(targ, 4, utilSecToFrames(3, 1)) -- 3 sec ENCHANT_SLOWED
    end -- don't slow allies
    soundMake(772, x, y)
    unitSpeed(mob, vx, vy)
end

-- Mob skill handler tables
MobArena.UrchinShaman = {
    Attack = shamanAttack,
    Meteor = shamanSummonMeteor,
    Confuse = shamanApplyConfuse,
    HealRitual = shamanHealingRitual
    }
MobArena.MechFlyer = {
    TripleShot = mechFlyerTripleShot,
    SelfRepair = mechFlyerSelfRepair,
    Attack = defaultMissileAttack
    }
MobArena.Skeleton = {
    Attack = defaultMeleeAttack
    }
MobArena.Ogre = {
    Attack = defaultMeleeAttack,
    MagicResistance = ogreMagicResistance
    }
MobArena.Wasp = {
    Attack = defaultMeleeAttack,
    CreateNest = createWaspNest
    }
MobArena.Zombie = {
    Attack = defaultMeleeAttack,
    PlaceTrap = zombiePlaceTrap
    }
MobArena.Dryad = {
    Attack = dryadAttack,
    NWEffect = dryadNatureWrath,
    ApplyFear = dryadApplyFear
    }
MobArena.Scorpion = {
    Attack = defaultMeleeAttack
    }
MobArena.Gargoyle = {
    Attack = defaultMissileAttack,
    DeathShot = gargCreateDeathShot,
    PowerShot = gargCreatePowerShot
    }
MobArena.Archmage = {
    Attack = archmageAttack,
    ForceField = archmageCastFF,
    Restoration = archmageMassHeal,
    Immolation = archmageFlame
    }
MobArena.Necromancer = {
    Attack = necromancerAttack,
    ChillTouch = necromancerColdTouch,
    DeepFreeze = necroDeepFreeze,
    RaiseZombie = necromancerRaiseZombie
    }
MobArena.Troll = {
    Attack = defaultMeleeAttack,
    Harpoon = trollLaunchHarpoon,
    Stink = trollStinkEffect
    }
MobArena.Wolf = {
    Attack = defaultMeleeAttack
    }
MobArena.Spider = {
    Attack = defaultMeleeAttack,
    Poison = spiderRemotePoison,
    Spiderlings = spiderCallSpiderlings,
    ShootWeb = spiderShootWeb
    }
MobArena.Imp = {
    Attack = defaultMissileAttack,
    Clone = impCloneSelf,
    Haste = impApplyHaste,
    ClonesAttack = impClonesAttack
    }
MobArena.Bear = {
    Attack = defaultMeleeAttack,
    Bash = bearDoBash
    }
MobArena.Shade = {
    Attack = defaultMeleeAttack,
    Jump = shadeJump
    }