-- unitUtil.lua
-- by KirConjurer (angrykirc)

local intToPtr;
do
	local t = memAlloc(4)
	function intToPtr(num)
        	setPtrInt(t, 0, num)
        	return getPtrPtr(t, 0)
    	end
end -- thanks to Dio

local emptyUserdata = intToPtr(1)
local noxCosTable = intToPtr(0x5B6658)
local noxSinTable = intToPtr(0x5B665C)
local float255 = intToPtr(0x437F0000)
local gameFPSAddr = intToPtr(0x85B3FC)

hostPlayer = getPtrPtr(ptrCall2(0x417090, intToPtr(31), emptyUserdata), 0x808)

function unitIsMonster(unit)
    if (bitAnd(getUnitClass(unit), 2) == 2 and bitAnd(unitFlags(unit), 0x8000) ~= 0x8000) then return true end
    return false -- true only if unit class is MONSTER and flag DEAD is not set
end

function unitIsPlayer(unit)
    return bitAnd(getUnitClass(unit), 4) == 4
end

function unitIsDead(unit)
    if not objectExists(unit) then return true end
    return bitAnd(unitFlags(unit), 0x8000) == 0x8000
end

function unitFindParentPlayer(unit)
    return ptrCall2(0x4EC580, unit, unit)
end

function utilGetGameFPS()
    return getPtrInt(gameFPSAddr, 0)
end

function utilSecToFrames(seconds, multiplier)
    local val = utilGetGameFPS()
    val = val * seconds * multiplier
    return math.ceil(val)
end

local function memFree2(mem)
    ptrCall2(0x40425D, mem, mem)
end

function playerAddGold(plr, gold)
    return ptrCall2(0x4FA590, plr, intToPtr(gold))
end

function unitGetParent(unit)
    return getPtrPtr(unit, 0x1FC)
end

function netGreenLightning(x1, y1, x2, y2, etime)
    local rectStruct = memAlloc(16)
    setPtrInt(rectStruct, 0, x1)
    setPtrInt(rectStruct, 4, y1)
    setPtrInt(rectStruct, 8, x2)
    setPtrInt(rectStruct, 12, y2)
    ptrCall2(0x523790, rectStruct, intToPtr(etime))
    memFree2(rectStruct)
end

function creatureActionRoam(unit)
    local h = unitSetAction(unit, 10)
    setPtrInt(h, 4, 0)
    setPtrInt(h, 12, 0x80)
    return h
end

function unitLastDamageDiff(unit)
    local ldt = getPtrInt(unit, 0x218)
    return getFrameCounter() - ldt
end

function unitAddHP(unit, hp)
    return ptrCall2(0x4EE460, unit, intToPtr(hp))
end

function buffSetPower(unit, buff, pwr)
    setPtrByte(unit, 0x198 + buff, pwr)
    return unit
end

-- unitHP returns dword...
function unitGetHPSafe(unit)
    local hd = getPtrPtr(unit, 0x22C)
    return getPtrShort(hd, 0)
end

function unitTeleportWake(x, y, tx, ty, ftime)
    local t = createObject("TeleportWake", x, y)
    setPtrInt(t, 0x88, getFrameCounter() + ftime) 
    local c = memAlloc(8)
    setPtrFloat(c, 0, tx)
    setPtrFloat(c, 4, ty)
    setPtrPtr(t, 0x2BC, c)
    return t
end

function unitAllocNewHP(unit, hp)
    local hd = memAlloc(0x14)
    setPtrShort(hd, 0, hp)
    setPtrShort(hd, 2, hp)
    setPtrShort(hd, 4, hp)
    setPtrPtr(unit, 0x22C, hd)
    return hd
end

function unitGetMaxHP(unit)
    local hd = getPtrPtr(unit, 0x22C)
    return getPtrShort(hd, 4)
end

function unitSetMaxHealth(npc, val)
    local hd = getPtrPtr(npc, 0x22C)
    setPtrShort(hd, 0, val)
    setPtrShort(hd, 2, val)
    setPtrShort(hd, 4, val)
    return
end

-- sets/unsets NO_COLLIDE flag
function unitNoCollide(unit, collide)
    if (collide) then setPtrInt(unit, 0x10, bitXor(unitFlags(unit), 0x40))
    else setPtrInt(unit, 0x10, bitOr(unitFlags(unit), 0x40)) end
    return collide
end

function objectExists(unit)
    if not unit then return false end
    if getThingType(unit) == 0xACAC then return false end
    return true
end

-- sets/unsets AIRBORNE flag
function unitFlAirborne(unit, flag)
    if (flag) then setPtrInt(unit, 0x10, bitXor(unitFlags(unit), 0x4000))
    else setPtrInt(unit, 0x10, bitOr(unitFlags(unit), 0x4000)) end
    return flag
end

function createHarpoon(unit)
    -- works only for players
    return ptrCall2(0x538890, unit, emptyUserdata)
end

local function noxStr( str ) -- by Dio
 local size = #str
 local mem = memAlloc( size )
 setPtrByte( mem, size, 0 )
 for i=1, size do
  setPtrByte( mem, i-1, string.byte( str, i ) )
 end
 return mem
end

function findAudioID(audname)
    local str = noxStr(audname)
    return ptrCall2(0x40AF50, str, emptyUserdata)
end

function monsterGetSoundset(unit)
    return ptrCall2(0x424300, unit, unit)
end

function unitCanSeeUnit(unitA, unitB)
    local ax, ay = unitPos(unitA)
    local bx, by = unitPos(unitB)
    if (mapTraceRay(ax, ay, bx, by, 1) and not unitTestBuff(unitB, 0)) then
        return true
    end -- also check for invisibility
    return false
end

function unitGetLookAngle(unit)
    return getPtrShort(unit, 0x7C)
end

function arrowSetDamage(unit, dmg) -- works only for MonsterArrowCollide
    local collData = getPtrPtr(unit, 0x2BC)
    setPtrInt(collData, 0, dmg) -- SOLO
    setPtrInt(collData, 4, dmg) -- ARENA
    return unit
end

-- unitBecomePet for non-monsters
function unitSetParent(unitA, unitB)
    return ptrCall2(0x4EC290, unitB, unitA)
end

function unitNeedSync(unit)
    return ptrCall2(0x4E44F0, unit, unit)
end

function unitIsPoisonImmune(unit)
    local subclass = getPtrInt(unit, 0xC)
    return bitAnd(subclass, 0x200) == 0x200
end

function unitNoPoisonResist(unit)
    local subclass = getPtrInt(unit, 0xC)
    setPtrInt(unit, 0xC, bitXor(subclass, 0x200)) -- IMMUNE_POISON
    return unit
end

function unitSetLookAngle(unit, val)
    setPtrShort(unit, 0x7C, val)
    setPtrShort(unit, 0x7E, val)
    return unit
end

function utilCosFromDirect(dir)
    if (dir < 0) then dir = 255 + dir end
    if (dir > 255) then dir = dir - 255 end
    return getPtrFloat(noxCosTable, 2 * dir * 4)
end

function utilSinFromDirect(dir)
    if (dir < 0) then dir = 255 + dir end
    if (dir > 255) then dir = dir - 255 end
    return getPtrFloat(noxSinTable, 2 * dir * 4)
end

function playerGetName(plr)
    if (not plr) then return "<404>" end
    return playerInfo(plr).name
end
    
function netReportDequip(item)
    return ptrCall2(0x4D84C0, intToPtr(255), item)
end

function mobSetSpell(npc, spell, useflag)
    local uc = getPtrPtr(npc, 0x2EC)
    local offset = 4 * spell + 0x5D0
    setPtrInt(uc, offset, useflag)
    return
end

-- returns true if unit is not null and active (enabled)
function unitIsOn(unit)
    if (not unit) then return false end
    if (bitAnd(unitFlags(unit), 0x1000000) ~= 0x1000000) then return false end
    return true
end

function playerSubGold(plr, gold)
    return ptrCall2(0x4FA5D0, plr, intToPtr(gold))
end

function buffCancel(unit, buff)
    return ptrCall2(0x4FF5B0, unit, intToPtr(buff))
end

function playerGetGold(plr)
    local uc = getPtrPtr(plr, 0x2EC)
    local pi = getPtrPtr(uc, 0x114)
    return getPtrInt(pi, 0x874)
end

function netArrowTrapFX(x, y, n)
    local p = memAlloc(8) -- point struct
    setPtrFloat(p, 0, x)
    setPtrFloat(p, 4, y)
    ptrCall2(0x5238A0, p, intToPtr(n))
    memFree2(p)
    return true
end

function netGeneratorBreakFX(x, y)
    local p = memAlloc(8) -- point struct
    setPtrFloat(p, 0, x)
    setPtrFloat(p, 4, y)
    ptrCall2(0x523200, p, intToPtr(0xC8))
    memFree2(p)
    return true
end

function sendEarthquake(x, y, power)
    local p = memAlloc(8) -- point struct
    setPtrFloat(p, 0, x)
    setPtrFloat(p, 4, y)
    ptrCall2(0x4D9110, p, intToPtr(power))
    memFree2(p)
    return true
end

function mobCanMove(unit)
    return not (unitTestBuff(unit, 5) or unitTestBuff(unit, 25) or unitTestBuff(unit, 3) or unitTestBuff(unit, 11))
end

function createMagicMissile(x, y, parent)
    if (not parent) then return nil end
    local m = createObject("MagicMissile", x, y)
    local uc = getPtrPtr(m, 0x2EC)
    setPtrPtr(uc, 0, parent)
    unitSetParent(m, parent)
    return m
end

function createFallingMeteor(x, y, fallslow, damage)
    local m = createObject("Meteor", x, y)
    local uc = getPtrPtr(m, 0x2EC)
    setPtrInt(uc, 0, damage) -- meteor max damage
    setPtrInt(m, 0x14, bitOr(getPtrInt(m, 0x14), 0x20)) -- dunno
    ptrCall2(0x4E46F0, m, float255) -- unitRaise
    setPtrFloat(m, 0x6C, fallslow) -- negative values make falling faster, positive - slower
    return m
end

function createToxicCloud(x, y, lifetime)
    local tc = createObject("ToxicCloud", x, y)
    local uc = getPtrPtr(tc, 0x2EC)
    setPtrInt(uc, 0, lifetime)
    return tc
end

local function getSomeRewardData()
    return ptrCall2(0x4E3CC0, emptyUserdata, emptyUserdata)
end

function activateRewardMarker(marker)
    return ptrCall2(0x4F0720, marker, getSomeRewardData())
end

function unitSetOn(unit)
	return ptrCall2(0x4E75B0, unit, emptyUserdata)
end

function unitSetOff(unit)
	return ptrCall2(0x4E7600, unit, emptyUserdata)
end

function unitCanAttack(unitA, unitB)
    local p = ptrCall2(0x5330C0, unitA, unitB)
    if (p == emptyUserdata) then return true end
    return false
end

function unitSetTeam(unit, team)
    if (unit) then
        setPtrInt(unit, 0x34, team)
        --setTimeout(ptrCall2(0x004E44F0, unit, unit)
    end
end

function createTreasure(x, y)
    local tre = createObject("TreasurePile", x, y)
    setPtrInt(tre, 0x2C4, 0x4F3580) -- TreasurePickup
    setPtrInt(tre, 0x2C8, 0x4ED710) -- TreasureDrop
    return tre
end -- use with gameFlags(bitOr(gameFlags(), 64))

function unitGetTeam(unit)
    if (unit) then
        return getPtrInt(unit, 0x34)
    end
end

function unitDamage(unit, val)
    setPtrInt(unit, 0x218, getFrameCounter()) -- damage taken
    return ptrCall2(0x4EE5E0, unit, intToPtr(val))
end

function unitReanimate(unit)
    if (unit) then
        unitActionPop(unit)
        unitActionPop(unit)
        unitSetAction(unit, 0) -- ACTION_IDLE
        local f = getPtrInt(unit, 0x10)
        f = bitXor(f, 0x8000) -- flDead
        f = bitXor(f, 0x18) -- ???
        unitHP(unit, 10)
        setPtrInt(unit, 0x10, f)
        soundMake(764, unitPos(unit))
    end
end

function unitPoison(unit, lvl)
    if (unit) then
        if (lvl > 0) then
            soundMake(100, unitPos(unit))
            setPtrByte(unit, 0x21C, lvl)
            --[[setPtrShort(unit, 0x21E, 1000)
            local hpData = getPtrPtr(unit, 0x22C)
            setPtrInt(hpData, 0x10, getFrameCounter())
            if (bitAnd(getUnitClass(unit), 4)) then
                local uc = getPtrPtr(unit, 0x2EC)
                local pd = getPtrPtr(uc, 0x114)
                local fl = getPtrInt(pd, 0xE60)
                setPtrInt(pd, 0xE60, bitOr(fl, 0x400))
            end]]--
        else
            setPtrByte(unit, 0x21C, 0)
            --setPtrShort(unit, 0x21E, 0)
        end
        ptrCall2(0x4EEA90, unit, intToPtr(lvl)) 
        -- temporary hack: call original function which sends update packet
    end
end