-- MobArena pre-loader

MobArena = { } -- create table

-- load main script file
dofile('.\\scripts\\mobarena.lua')

-- Team spawn coordinates
MobArena.SpawnLocations = { { 1943, 1805 }, { 4128, 4174 } }
-- Structure coordinates (global)
MobArena.BaseLocations = { { 2530, 1633 }, { 3520, 4370 } }

MobArena.GetSpawnCoords = function(team)
    return MobArena.SpawnLocations[team][1], MobArena.SpawnLocations[team][2]
end

local function onStructureDestroyed(unit, attacker)
    local team = unitGetTeam(attacker)
    if team then
        teamScore(team, 100)
    end
    return false
end

local lastChatAlarm = 0 -- prevents console spam

local function onStructureDamaged(damaged, parent, damager, value, typ)
    local team = unitGetTeam(damaged)
    if not parent then return false, 0 end
    local pplayer = unitFindParentPlayer(parent)
    if not pplayer then return false, 0 end
    local attTeam = unitGetTeam(pplayer)

    if attTeam == 0 then return false, 0 end
    if value > 1 then
        netPointFx(0x8B, unitPos(damaged))
    end
    
    if (MobArena.BaseDamagedHandler) then
        MobArena.BaseDamagedHandler(team)
    end -- used by AI
    
    if (getFrameCounter() - lastChatAlarm) > 30 then
        if team ~= attTeam then
            lastChatAlarm = getFrameCounter()
            local msg = "Base is under attack! HP: "
            local hp = unitGetHPSafe(damaged) - value
            if hp < 0 then hp = 0 end
            msg = msg .. hp
            sendChat(-1, msg, servNetCode(damaged), false)
        end
    end
    return false
end

local function createStructure(x, y, team, name)
    local u = createObject(name, x, y)
    --setPtrByte(getPtrPtr(u, 0x2EC), 0x5C, 0) -- no spawning
    unitAllocNewHP(u, 500) -- 500 'durability points'
    unitSetTeam(u, team)
    --[[local flags = unitFlags(u)
    flags = bitXor(flags, 0x10) -- this allows melee attacking
    setPtrInt(u, 0x10, flags)]]--
    unitOnDamage(u, onStructureDamaged)
    unitOnDie(u, onStructureDestroyed)
    return u
end

-- Structure setup
createStructure(MobArena.BaseLocations[1][1], MobArena.BaseLocations[1][2], 1, "MovableStatueVictory5SW")
createStructure(MobArena.BaseLocations[2][1], MobArena.BaseLocations[2][2], 2, "MovableStatueVictory5NE")
-- End setup

-- Mob spawning functions

local function getSpawnPos(plr)
    local team = unitGetTeam(plr)
    if (team ~= 1 and team ~= 2) then
        print("You need to create teams before playing this map.")
    else
        return MobArena.GetSpawnCoords(team)
    end
    return nil
end

local function showTutorial(plr)
    local plid = playerInfo(plr).idx
    sendChat(plid, "Use mouse to move/attack. Press J/K/L/Space to use monster's skills; F11 to hide GUI.", servNetCode(plr), false)
    return true
end

function spawnShamanP()
    local plr = scriptParent
    local x, y = getSpawnPos(plr)
    sendChat(-1, "Urchin Shaman!", servNetCode(plr), false)
    showTutorial(plr)
    spawnShaman(x, y, plr)
    return
end

function spawnMechFlyerP()
    local plr = scriptParent
    local x, y = getSpawnPos(plr)
    sendChat(-1, "Galava Flyer!", servNetCode(plr), false)
    showTutorial(plr)
    spawnMechFlyer(x, y, plr)
    return
end

function spawnSkeletonP()
    local plr = scriptParent
    local x, y = getSpawnPos(plr)
    sendChat(-1, "Skeleton!", servNetCode(plr), false)
    showTutorial(plr)
    spawnSkeleton(x, y, plr)
    return
end

function spawnOgreP()
    local plr = scriptParent
    local x, y = getSpawnPos(plr)
    sendChat(-1, "Ogre!", servNetCode(plr), false)
    showTutorial(plr)
    spawnOgre(x, y, plr)
    return
end

function spawnZombieP()
    local plr = scriptParent
    local x, y = getSpawnPos(plr)
    sendChat(-1, "Zombie!", servNetCode(plr), false)
    showTutorial(plr)
    spawnZombie(x, y, plr)
    return
end

function spawnWaspP()
    local plr = scriptParent
    local x, y = getSpawnPos(plr)
    sendChat(-1, "Wasp!", servNetCode(plr), false)
    showTutorial(plr)
    spawnWasp(x, y, plr)
    return
end

function spawnDryadP()
    local plr = scriptParent
    local x, y = getSpawnPos(plr)
    sendChat(-1, "Dryad!", servNetCode(plr), false)
    showTutorial(plr)
    spawnDryad(x, y, plr)
    return
end

function spawnScorpionP()
    local plr = scriptParent
    local x, y = getSpawnPos(plr)
    sendChat(-1, "Scorpion!", servNetCode(plr), false)
    showTutorial(plr)
    spawnScorpion(x, y, plr)
    return
end

function spawnArchmageP()
    local plr = scriptParent
    local x, y = getSpawnPos(plr)
    sendChat(-1, "Archmage!", servNetCode(plr), false)
    showTutorial(plr)
    spawnArchmage(x, y, plr)
    return
end

function spawnGargoyleP()
    local plr = scriptParent
    local x, y = getSpawnPos(plr)
    sendChat(-1, "Gargoyle!", servNetCode(plr), false)
    showTutorial(plr)
    spawnGargoyle(x, y, plr)
    return
end

function spawnShadeP()
    local plr = scriptParent
    local x, y = getSpawnPos(plr)
    sendChat(-1, "Shade!", servNetCode(plr), false)
    showTutorial(plr)
    spawnShade(x, y, plr)
    return
end

function spawnBearP()
    local plr = scriptParent
    local x, y = getSpawnPos(plr)
    sendChat(-1, "Bear!", servNetCode(plr), false)
    showTutorial(plr)
    spawnBear(x, y, plr)
    return
end

function spawnImpP()
    local plr = scriptParent
    local x, y = getSpawnPos(plr)
    sendChat(-1, "Imp!", servNetCode(plr), false)
    showTutorial(plr)
    spawnImp(x, y, plr)
    return
end

function spawnSpiderP()
    local plr = scriptParent
    local x, y = getSpawnPos(plr)
    sendChat(-1, "Spider!", servNetCode(plr), false)
    showTutorial(plr)
    spawnSpider(x, y, plr)
    return
end

function spawnWolfP()
    local plr = scriptParent
    local x, y = getSpawnPos(plr)
    sendChat(-1, "Wolf!", servNetCode(plr), false)
    showTutorial(plr)
    spawnWolf(x, y, plr)
    return
end

function spawnTrollP()
    local plr = scriptParent
    local x, y = getSpawnPos(plr)
    sendChat(-1, "Troll!", servNetCode(plr), false)
    showTutorial(plr)
    spawnTroll(x, y, plr)
    return
end

function spawnNecroP()
    local plr = scriptParent
    local x, y = getSpawnPos(plr)
    sendChat(-1, "Necromancer!", servNetCode(plr), false)
    showTutorial(plr)
    spawnNecromancer(x, y, plr)
    return
end